import time
import math

# MIT License
# https://github.com/chrischoy/SpatioTemporalSegmentation/blob/master/lib/utils.py
class Timer(object):
    """A simple timer."""

    def __init__(self):
        self.reset()

    def reset(self):
        self.total_time = 0
        self.calls = 0
        self.start_time = 0
        self.diff = 0
        self.average_time = 0

    def tic(self):
        self.start_time = time.time()

    def toc(self, average=True):
        self.diff = time.time() - self.start_time
        self.total_time += self.diff
        self.calls += 1
        self.average_time = self.total_time / self.calls

        return self.average_time if average else self.diff

    @property
    def avg(self):
        return self.average_time

    def __repr__(self):
        return f"Timer(average time: {self.avg})"

# MIT License
# https://github.com/chrischoy/SpatioTemporalSegmentation/blob/master/lib/utils.py
class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def stats(self):
        return {
            'avg' : self.avg,
            'count' : self.count,
            'val' : self.val,
        }

    def __repr__(self):
        return f"AverageMeter(avg: {self.avg}, count: {self.count})"

class TotalMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.values = []

    def update(self, val):
        self.values.append(val)

    @property
    def count(self):
        return len(self.values)

    @property
    def avg(self):
        return np.mean(self.values)

    @property
    def std(self):
        return np.std(self.values)

    @property
    def var(self):
        return np.var(self.values)

    def stats(self):
        return {
            'avg' : self.avg,
            'count' : self.count,
            'std' : self.std,
            'var' : self.var,
        }

    def __repr__(self):
        return f"TotalMeter(avg: {self.avg}, std: {self.std}, count: {self.count})"

# https://github.com/mlb2251/mlb
class ProgressBar:
    def __init__(self, num_steps, num_dots=10):
        self.num_steps = num_steps
        self.num_dots = num_dots
        self.curr_step = 0
        self.dots_printed = 0

    def step(self):
        expected_dots = math.ceil(self.curr_step / self.num_steps * self.num_dots)
        dots_to_print = expected_dots - self.dots_printed
        if dots_to_print > 0:
            print('.'*dots_to_print, end='', flush=True)
        self.dots_printed = expected_dots
        self.curr_step += 1
        if self.curr_step == self.num_steps:
            print('!\n', end='', flush=True)


def timeit(f, n=5, verbose=False):
    @functools.wraps(f)
    def foo(*args, **kwargs):
        meter = TotalMeter()

        for _ in range(n):
            t = time.time()
            f(*args, **kwargs)
            meter.update(time.time() - t)

        if verbose:
            print(f"{meter.avg} ms ± {meter.std} (mean ± std over {n} runs)")

        return meter.avg, meter.std
    
    return foo


from colorama import Fore, Back, Style

def color(s,c):
    return '\001'+c+'\002'+str(s)+'\001'+Style.RESET_ALL+'\002'

""" Add colors to a string and return the string """

def green(s):
    return color(s, Fore.GREEN)
    
def red(s):
    return color(s, Fore.RED)

def pink(s):
    return color(s, Fore.MAGENTA)

def blue(s):
    return color(s, Fore.BLUE)

def cyan(s):
    return color(s, Fore.CYAN)

def yellow(s):
    return color(s, Fore.YELLOW)

def gray(s):
    return color(s,'\033[90m')


""" Add style to a string (bold, underline) """

def bold(s):
    return color(s,'\033[1m')

def underline(s):
    return color(s,'\033[4m')



import os

def read(filename):
    """Reads the data in a given file."""

    with open(filename, 'r') as f:
        return f.read()

def readlines(path):
    """Reads the lines in a given file."""

    with open(path, 'r') as f:
        return f.read().splitlines()

def listdir(path):
    """Returns a list of files in the given directory with the base path name included."""

    return [os.path.join(path, f) for f in os.listdir(path)]

def write(path, string):
    """Writes a given string to a file in append mode."""

    with open(path, 'a') as f:
        f.write(string)


import os, sys
from bdb import BdbQuit
from contextlib import contextmanager
from pdb import post_mortem
import traceback
import types
import subprocess
import logging

import jaustin
import jaustin.color

""" Command utilities """

def run(command, timeout=120):
    """Run a given shell command in a subprocess and return the result with captured stdout and stderr."""

    result = subprocess.run(command.split(" "), stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=timeout)
    return result


def decompress(filename, timeout=120):
    """Decompress a given .rar or .zip file to the same root directory. Requires
       unrar and unzip to be installed."""

    assert os.path.isfile(filename), "File does not exist or is not a valid file."

    dirname = os.path.dirname(filename)
    _, ext = os.path.splitext(os.path.basename(filename))

    ext_dict = {  # two spaces as a hack to handle escaped filenames
        '.rar' : 'unrar  -o+  x  {filename} {dirname}',
        '.zip' : 'unzip  -o  {filename}  -d  {dirname}'
    }

    if ext not in ext_dict.keys():
        raise TypeError(f"Unable to decompress file with unknown extension {ext}.")
    
    command = ext_dict[ext].format(filename=filename, dirname=dirname)
    logging.info(f"Extracting {ext} archive (running ({command.split('  ')})")
    result = jaustin.run(command, timeout=timeout)
    logging.info(f"Extraction completed with return code {result.returncode}.")

    if result.returncode != 0:
        print("[WARNING]", result.stderr)

    return result
        

""" Examination utilities """

def help(foo):
    """Prints any documentation attached to a given method."""

    print(foo.__doc__)

def examine(obj):
    """Examines a given object's attributes and print them color-coded by type."""
    
    fields = []
    methods = []
    for entry in dir(obj):
        try:
            value = getattr(obj, entry) 
        except:
            continue
        if type(value) is types.BuiltinMethodType or type(value) is type or type(value) is types.MethodType or type(value) is types.MethodWrapperType:
            methods.append(entry)
        else:
            fields.append(entry)

    print(jaustin.color.bold("==========  Summary  ==========\n"))

    lspace = ""

    attributes = {
        'type' : obj.__class__.__name__,
        'no. methods' : len(methods),
        'no. fields' : len(fields),
        'help' : obj.__doc__,
    }

    for field, value in attributes.items():
        print(f"{lspace}{jaustin.color.green(field.ljust(15))} : {jaustin.color.cyan(value)}")

    print()

    print(f"{lspace}{jaustin.color.blue('methods:')}")
    for entry in methods:
        print("        " + jaustin.color.cyan(entry))

    print()

    print(f"{lspace}{jaustin.color.blue('fields:')}")
    for entry in fields:
        print("        " + jaustin.color.cyan(entry))

@contextmanager
def debug(do_debug=True):
    """Context manager that ddrops into the PDB post_mortem debugger on failure.

    Example:
        >>> with jaustin.debug():
        >>>     main()
    """

    try:
        yield None
    except BdbQuit:
        raise
    except KeyboardInterrupt:
        raise
    except Exception as e:
        if do_debug:
            print(''.join(traceback.format_exception(e.__class__,e,e.__traceback__)))
            print(e)
            post_mortem()
            sys.exit(1)
        else:
            raise e
    finally:
        pass


import functools

def methodcache(f):
    """Utility decorator that ensures the passed function is only called once and the result 
    is cached and reused on future calls. This saves the cached values in the method __cache
    attribute (only works for methods).
    """

    name = f.__name__

    @functools.wraps(f)
    def cached(self, *args, **kwargs):
        if not hasattr(self, "__cache"):
            self.__cache = {}

        if name not in self.__cache:
            self.__cache[name] = {"called": False, "result": None}

        if self.__cache[name]["called"]:
            return self.__cache[name]["result"]

        result = f(self, *args, **kwargs)
        self.__cache[name] = {"called": True, "result": result}
        return result

    return cached




import os

def picklecache(match_args=True, reset=False):
    """Caching method that pickles arguments and return values for a given
    function and saves them for reuse across multiple runs of the same program.

    Args:
        match_args (bool, optional): if True, only returns the cached value if 
            args and kwargs match those saved. Otherwise, returns the value regardless.

        reset (bool, optional): if True, resets the cache and doesn't load saved values.
    """

    def real_decorator(f):
        import pickle
        import functools

        def dump(path, args, kwargs, data):
            with open(path, "wb") as f:
                d = {"args" : args, "kwargs" : kwargs, "data" : data}
                pickle.dump(d, f, pickle.HIGHEST_PROTOCOL)
        
        if not os.path.exists(".picklecache"):
            os.mkdir(".picklecache")

        if reset:
            path = os.path.join(".picklecache", "{}".format(f.__name__))

            if os.path.exists(path):
                os.remove(path)
        
        @functools.wraps(f)  
        def wrapper(*args, **kwargs):
            path = os.path.join(".picklecache", "{}".format(f.__name__))

            if os.path.exists(path):
                try:
                    with open(path, "rb") as pfile:
                        cache = pickle.load(pfile)

                    if match_args:
                        cached_args = cache["args"]
                        cached_kwargs = cache["kwargs"]

                        if args == cached_args and kwargs == cached_kwargs:
                            print("Loading data from cache for function {}".format(f.__name__))
                            return cache["data"]

                        else:
                            data = f(*args, **kwargs)
                            dump(path, args, kwargs, data)
                            return data
                    else:
                        print("Loading data from cache for function {}".format(f.__name__))
                        return cache["data"]

                except Exception as e:
                    print("Corrupted pickle cache. Resetting cache for function {}".format(f.__name__))
                    if os.path.exists(path):
                        os.remove(path)

                    retval = f(*args, **kwargs)
                    dump(path, args, kwargs, retval)
                    return retval

            else:
                retval = f(*args, **kwargs)
                dump(path, args, kwargs, retval)
                return retval
        return wrapper

    return real_decorator


# Copyright (c) NVIDIA Corporation. All rights reserved
#
# TrainerBase is modified from the WeightedProcrustesTrainer from https://github.com/chrischoy/DeepGlobalRegistration/blob/master/core/trainer.py
# Copyright (c) Chris Choy (chrischoy@ai.stanford.edu) and Wei Dong (weidong@andrew.cmu.edu), MIT License
import os
import logging
import numpy as np
import json
import math
from typing import Union
from datetime import datetime
import sys

from torch.utils.tensorboard import SummaryWriter
from torch.utils.data import DataLoader

eps = np.finfo(float).eps

class TrainerBase:
    def __init__(
        self,
        data_loader: DataLoader,
        val_data_loader: DataLoader,
        device,
        checkpoint: str = None,
        cfg: dict = None,
    ):
        logging.info("Initializing a trainer.")

        self.cfg = cfg

        # initialize logging
        self.checkpoint_dir = self.get_checkpoint_dir()

        logdir = os.path.join(checkpoint_dir, "debug.log")
        logging.root.handlers = []

        handlers = [
            logging.StreamHandler(sys.stdout),
            logging.FileHandler(logdir)
        ]

        logging.basicConfig(
            level=logging.root.level,
            format="%(asctime)s [%(levelname)s] %(message)s",
            handlers=handlers,
        )

        logging.info(f"saving log files to {logdir}.")

        self.writer = SummaryWriter(self.checkpoint_dir)
        
        # Training config and logging
        self.start_epoch, self.train_iters, self.train_steps_per_epoch = self.get_training_info()

        epoch_steps = self.train_steps_per_epoch
        self.max_epoch = int(math.ceil(self.train_iters / epoch_steps))

        # Data loader
        self.data_loader = data_loader
        self.val_data_loader = val_data_loader

        self.train_data_loader_iter = iter(self.data_loader)
        self.test_valid = True if self.val_data_loader is not None else False
        self.device = device

        # Validation config
        self.validate_every, self.best_metric, self.best_val = self.get_valid_info()
        self.best_val_epoch = -np.inf

        self.initialize_model()

        self.initialize_optimizer()
        self.initialize_scheduler()

        if checkpoint is not None:
            self.load_checkpoint(checkpoint)

        self.dump_config()

    def get_checkpoint_dir(self) -> str:
        """
        Return the checkpoint directory for the current run. For example,
        can return self.cfg.experiment_dir.
        """

        raise NotImplementedError

    def initialize_logging(self, checkpoint_dir)
        logdir = os.path.join(checkpoint_dir, "debug.log")
        logging.root.handlers = []

        handlers = [
            logging.StreamHandler(sys.stdout),
            logging.FileHandler(logdir)
        ]

        logging.basicConfig(
            level=logging.root.level,
            format="%(asctime)s [%(levelname)s] %(message)s",
            handlers=handlers,
        )

        logging.info(f"saving log files to {logdir}.")

    def initialize_distributed_training(self, rank) -> None:
        raise NotImplementedError

    def initialize_model(self) -> None:
        raise NotImplementedError
        # Something like this
        ModelClass = load_model(self.cfg)
        self.model_a = ModelClass(A, B, C, args).to(self.device)
        self.model_b = ModelClass(C, D, E, args).to(self.device)

    def get_training_info(self):
        """get_training_info should return self.start_epoch, self.train_iters, 
        self.train_steps_per_epoch. for example, can return self.cfg.train.start_epoch, 
        self.cfg.train.iters, self.cfg.train.train_steps_per_epoch."""

        raise NotImplementedError

    def get_valid_info(self):
        """get_val_info should return self.validate_every, self.best_metric, self.init_metric,
        where validate_every determines how many epochs to validate after, best_metric is the string
        name for the best metric returned by _valid_epoch, and init_mtric is the initial value."""
        raise NotImplementedError

        # return name of metric and initial value
        return self.cfg.valid.validate_every, "psnr", -np.inf
        

    def initialize_optimizer(self) -> None:
        raise NotImplementedError
        # Loss and optimizer
        config = self.cfg
        self.param_optimizer = getattr(optim, config.optimizer)(
            self.network.parameters(),
            lr=config.lr,
            momentum=config.momentum,
            weight_decay=config.weight_decay,
        )

    def initialize_scheduler(self) -> None:
        raise NotImplementedError

    def load_checkpoint(self, checkpoint) -> None:
        raise NotImplementedError

    def dump_config(self):
        json.dump(
            cfg,
            open(os.path.join(self.checkpoint_dir, "config.json"), "w"),
            indent=4,
            sort_keys=False,
        )

    def pre_epoch(self, epoch):
        """called at the beginning of each epoch. For example, can run
        torch.cuda.empty_cache()."""

        return

    def train(self):
        """
        Full training logic: train, valid, and save
        """
        # Train and valid
        for epoch in range(self.start_epoch, self.max_epoch + 1):
            self.pre_epoch(epoch)

            # validate
            if self.test_valid and epoch % self.validate_every == 0:
                val_dict = self._valid_epoch(epoch)

                if self.best_val < val_dict[self.best_val_metric]:
                    self.info(
                        f"Saving the best val model with {self.best_val_metric}: {val_dict[self.best_val_metric]}",
                    )
                    self.best_val = val_dict[self.best_val_metric]
                    self.best_val_epoch = epoch
                    self._save_checkpoint(epoch, "best_val_checkpoint")

                else:
                    logging.info(
                        f"Current epoch: {epoch}. best val model with {self.best_val_metric}: {self.best_val} at epoch {self.best_val_epoch}"
                    )
                    self._save_checkpoint(epoch, "checkpoint")

            self.log_scheduler(epoch)
            self._train_epoch(epoch)
            self.step_scheduler(epoch)

    def eval(self, save_disparity_image=False):
        raise NotImplementedError

    def log_scheduler(self, epoch):
        lr = self.optimizer.param_groups[0]['lr']
        logging.info(f" Epoch: {epoch}, LR: {lr}")

        self.writer.add_scalar("train/learning_rate", lr, epoch)

    def step_scheduler(self, epoch):
        self.scheduler.step()

    def step_and_zero_optimizer(self, step):
        self.optimizer.step()
        self.optimizer.zero_grad()

    def _train_epoch(self, epoch: int) -> None:
        raise NotImplementedError

    def _valid_epoch(self) -> dict:
        raise NotImplementedError

    def _save_checkpoint(self, epoch: int, filename="checkpoint") -> None:
        raise NotImplementedError